from numpy import ndarray, array


def fix_slicing(index, logical_length, allow_column=True):
    """
    Checks, and caps, the slices being used.
    :param index: The initial index, which may be int or slice.
    :param logical_length: The current logical length of the data. It is used to cap the indices.
      If None, indices will not be capped.
    :param allow_column: Tells whether allowing 1 or 2 dimensions indexing.
    :return: The fixed slicing, if no exception occurs.
    """

    column_index = None
    if isinstance(index, tuple):
        if allow_column:
            if len(index) == 2:
                index, column_index = index
            else:
                raise IndexError("At most two indices/slices (bi-dimensional indexing) are supported")
        else:
            raise IndexError("Only one index/slice (one-dimensional indexing) is supported")

    if isinstance(index, slice):
        start = index.start
        stop = index.stop
        step = index.step
        if start is None:
            start = 0
        if stop is None:
            stop = logical_length
        if step and step != 1:
            raise IndexError("Slices with step != 1 are not supported")
        if start < 0 or (stop is not None and stop < 0):
            raise IndexError("Negative indices in slices are not supported")
        if stop is not None and stop < start:
            raise IndexError("Slices must have start <= stop indices")
        elif logical_length is None:
            return start, stop, column_index
        else:
            # Here, start will be <= stop. We will limit both by the logical_length, silently.
            return min(start, logical_length), min(stop, logical_length), column_index
    elif isinstance(index, int):
        if index < 0:
            raise IndexError("Negative indices are not supported")
        if logical_length is None:
            return index, None, column_index
        else:
            return min(index, logical_length), None, column_index
    else:
        raise TypeError("Only slices (non-negative, growing, and with step 1) or non-negative integer indices are "
                        "supported")


def fix_input(index, expected_width, expected_length, expected_type, value):
    """
    Checks and fixes the value according to the given index type.
    :param index: The initial index, which may be int or slice.
    :param expected_width: The expected width of the value.
    :param expected_length: The expected length of the data. It will be ignored if the index is a number.
    :param expected_type: The expected type for array input.
    :param value: The given value, which may be scalar or array.
    :return: The fixed value, if no exception occurs.
    """

    # If the input is an iterable, convert it to an 1-dimensional array.
    if isinstance(value, (tuple, list)):
        value = array(value)

    if isinstance(index, slice):
        if not isinstance(value, ndarray) and value.dtype != expected_type:
            raise TypeError("When setting a slice, the value must be a numpy array of (stop - start)x(width) "
                            "elements (if the width is 1, an uni-dimensional array of (stop-start) elements is "
                            "also allowed)")
        if expected_width == 1 and len(value.shape) == 1:
            value = value[:]
            value.shape = (value.size, 1)
        if value.shape != (expected_length, expected_width):
            raise TypeError("When setting a slice, the value must be a numpy array of (stop - start)x(width) "
                            "elements (if the width is 1, an uni-dimensional array of (stop-start) elements is "
                            "also allowed)")
    elif isinstance(index, int):
        if expected_width > 1 and (not isinstance(value, ndarray) or value.shape != (expected_width,)):
            raise TypeError("When setting an index, the value must be a 1-dimensional numpy array of the appropriate "
                            "size (=width), if the expected width is > 1")
        if expected_width == 1 and not isinstance(value, ndarray):
            value = array((value,))
    else:
        raise TypeError("Only slices (non-negative, growing, and with step 1) or non-negative integer indices are "
                        "supported")
    return value


def chunked_slicing(slice_start, slice_stop, chunk_size):
    """
    Iterator that yields, every time, a data structure like this:
      - Current chunk index
      - Start index in chunk
      - Stop index in chunk
      - Start index in source/destination
      - Stop index in source/destination

    It is guaranteed: 0 <= slice_start <= slice_stop <= logical length <= chunk_size * chunk_count.

    The algorithm goes like this:
      - Before: we know the starter bin, and the end bin.
    :param slice_start: The overall start.
    :param slice_stop: The overall stop.
    :param chunk_size: The chunk size.
    :return: A generator.
    """

    start_chunk = slice_start // chunk_size
    stop_chunk = slice_stop // chunk_size
    if start_chunk == stop_chunk:
        # This is the easiest case.
        # The data indices will be 0 and stop_chunk - start_chunk.
        # The chunk index will be start_chunk.
        # the start index in chunk, and end index in chunk, will both involve modulo.
        data_indices = (0, slice_stop - slice_start)
        chunk_indices = (slice_start % chunk_size, slice_stop % chunk_size)
        yield data_indices, start_chunk, chunk_indices
    else:
        # In this case, start chunk will always be lower than end chunk.
        chunk_start_index = slice_start % chunk_size
        chunk_stop_index = slice_stop % chunk_size
        first_iteration = True
        data_index = 0
        current_chunk = start_chunk
        # We know that, in the first iteration:
        # - chunk_start_index >= 0
        # - start_chunk < stop_chunk, strictly
        # And in further iterations:
        # - chunk_start_index == 0
        # - start_chunk <= stop_chunk
        while True:
            # The chunk upper bound will be:
            # - The chunk stop index, if in last chunk.
            # - The chunk size, otherwise.
            if current_chunk == stop_chunk:
                # Another check here: if the chunk stop index is 0, just break.
                if chunk_stop_index == 0:
                    return
                current_chunk_ubound = chunk_stop_index
            else:
                current_chunk_ubound = chunk_size
            # The chunk lower bound will be:
            # - The chunk start index, if in first chunk.
            # - 0, otherwise.
            if first_iteration:
                current_chunk_lbound = chunk_start_index
            else:
                current_chunk_lbound = 0
            # Now we know the chunk start and chunk end bounds.
            # We also know the current chunk index.
            # We also know the current data index.
            # Also we can compute the current data length
            length = current_chunk_ubound - current_chunk_lbound
            # We can yield all the data now.
            data_indices = (data_index, data_index + length)
            chunk_indices = (current_chunk_lbound, current_chunk_ubound)
            yield data_indices, current_chunk, chunk_indices
            # If this is the last chunk, we must exit.
            if current_chunk == stop_chunk:
                return
            # If not, then we move the data index and the current chunk.
            current_chunk += 1
            data_index += length
            # Finish the first iteration.
            first_iteration = False



from numpy import nan
from ..sources import Source
from ..utils.tail_runners import TailRunner
from ..utils.mappers.smart_pluckers import smart_plucker
from . import Indicator


class Slope(Indicator):
    """
    Computes the nominal difference in prices between the current instant and the previous one.
    For the instant 0, computes the nominal difference in prices between that instance and the
      initial value. If the initial value is None, the nominal difference will be NaN.
    Since time intervals are constant, these differences are, in turn, the change slopes.

    Two additional arguments: component and row. They are required depending on the given parent.
    """

    def __init__(self, parent, side=Source.ASK, component='end', row=0):
        self._parent = smart_plucker(parent, side, component, row)
        self._tail_runner = TailRunner(2)
        Indicator.__init__(self, parent)

    def _update(self, start, end):
        """
        Updates the indices with the moving/tailed mean for -respectively- each index.
        :param start: The start index to update.
        :param end: The end index to update.
        """

        for idx, chunk, incomplete in self._tail_runner.tail_iterate(start, end, self._parent):
            if incomplete:
                if self._parent.initial is None:
                    self._data[idx] = nan
                else:
                    self._data[idx] = chunk[0] - self._parent.initial
            else:
                self._data[idx] = chunk[1] - chunk[0]

    @property
    def parent(self):
        """
        The parent indicator.
        """

        return self._parent



from numpy import empty, nan, hstack
from .. import Indicator
from .mean import MovingMean


class MovingVariance(Indicator):
    """
    Based on a moving mean indicator, this indicator tracks the variance and/or the standard deviation.
    Aside from moving mean, the following arguments may be specified:
    - Variance: Include the variance in the computation.
    - Std. Error: Include the standard error in the computation.
    - Unbiased: Whether to use the unbiased sample variance instead of the natural (biased) one.
    """

    def __init__(self, moving_mean, var=False, stderr=True, unbiased=True):
        if not isinstance(moving_mean, MovingMean):
            raise TypeError("For MovingVariance instances, the only allowed source indicator is a moving mean")
        if not (var or stderr):
            raise ValueError("At least one of the `var` or `stderr` flags must be specified")
        self._use_var = var
        self._use_stderr = stderr
        self._with_unbiased_correction = unbiased
        self._moving_mean = moving_mean
        Indicator.__init__(self, moving_mean)

    def width(self):
        """
        We may use both flags here, so the width may be 2.
        """

        if self._use_var and self._use_stderr:
            return 2
        return 1

    def _update(self, start, end):
        """
        Adds calculation of the variance and/or
        :param start: The start index to update.
        :param end: The end index to update.
        :return:
        """

        means = self._moving_mean[start:end]
        tail_runner = self._moving_mean.tail_runner
        tail_size = tail_runner.tail_size
        n = tail_size
        if self._with_unbiased_correction:
            n -= 1

        # This one HAS to be calculated.
        variance = empty((end - start, 1), dtype=float)
        for idx, chunk, incomplete in tail_runner.tail_iterate(start, end, self._moving_mean.parent):
            mean = means[idx - start]
            if incomplete:
                variance[idx - start] = nan
            else:
                variance[idx - start] = ((chunk - mean) ** 2).sum() / n

        # If we need the standard error, we also have to calculate this one.
        stderr = None
        if self._use_stderr:
            stderr = variance ** 0.5

        # Now we must assign the data appropriately.
        if self._use_var and self._use_stderr:
            self._data[start:end] = hstack([variance, stderr])
        elif self._use_var:
            self._data[start:end] = variance
        elif self._use_stderr:
            self._data[start:end] = stderr



from enum import IntEnum


class Interval(IntEnum):
    """
    These values can be used as intervals for source (original data) or digested data.
    They can also constrain/truncate a timestamp to be relevant for the interval being considered.
    """

    SECOND = 1
    MINUTE = 60*SECOND
    MINUTE5 = 5*MINUTE
    MINUTE10 = 10*MINUTE
    MINUTE15 = 15*MINUTE
    MINUTE20 = 20*MINUTE
    MINUTE30 = 30*MINUTE
    HOUR = 60*MINUTE
    HOUR2 = 2*HOUR
    HOUR3 = 3*HOUR
    HOUR4 = 4*HOUR
    HOUR6 = 6*HOUR
    HOUR8 = 8*HOUR
    HOUR12 = 12*HOUR
    DAY = 24*HOUR

    def allowed_as_source(self):
        """
        Tells whether this interval can be used in a source frame.
        """

        return self != Interval.DAY

    def allowed_as_digest(self, for_source_interval=SECOND):
        """
        Tells whether this interval can be used in a digest frame given a related source frame size.
        Aside from being allowed, this interval size must be GREATER than the given, optional, one.
        :param for_source_interval: The source frame size to compare.
        """

        iself = int(self)
        ifor = int(for_source_interval)
        return iself % ifor == 0 and iself > ifor

    def round(self, stamp):
        """
        Rounds a timestamp down to the relevant interval. E.g. the MINUTE interval will round down,
          removing the seconds (setting them to 0), while the MINUTE5 will, also, round down the minutes
          in chunks of 5.
        :param stamp: The stamp to round down.
        :return: The rounded stamp.
        """
        if self == self.SECOND:
            return stamp.replace(microsecond=0)
        elif self == self.MINUTE:
            return stamp.replace(microsecond=0, second=0)
        elif self == self.MINUTE5:
            return stamp.replace(microsecond=0, second=0, minute=(stamp.minute // 5) * 5)
        elif self == self.MINUTE10:
            return stamp.replace(microsecond=0, second=0, minute=(stamp.minute // 10) * 10)
        elif self == self.MINUTE15:
            return stamp.replace(microsecond=0, second=0, minute=(stamp.minute // 15) * 15)
        elif self == self.MINUTE20:
            return stamp.replace(microsecond=0, second=0, minute=(stamp.minute // 20) * 20)
        elif self == self.MINUTE30:
            return stamp.replace(microsecond=0, second=0, minute=(stamp.minute // 30) * 30)
        elif self == self.HOUR:
            return stamp.replace(microsecond=0, second=0, minute=0)
        elif self == self.HOUR2:
            return stamp.replace(microsecond=0, second=0, minute=0, hour=(stamp.hour // 2) * 2)
        elif self == self.HOUR3:
            return stamp.replace(microsecond=0, second=0, minute=0, hour=(stamp.hour // 3) * 3)
        elif self == self.HOUR4:
            return stamp.replace(microsecond=0, second=0, minute=0, hour=(stamp.hour // 4) * 4)
        elif self == self.HOUR6:
            return stamp.replace(microsecond=0, second=0, minute=0, hour=(stamp.hour // 6) * 6)
        elif self == self.HOUR8:
            return stamp.replace(microsecond=0, second=0, minute=0, hour=(stamp.hour // 8) * 8)
        elif self == self.HOUR12:
            return stamp.replace(microsecond=0, second=0, minute=0, hour=(stamp.hour // 12) * 12)
        elif self == self.DAY:
            return stamp.replace(microsecond=0, second=0, minute=0, hour=0)
        else:
            raise AssertionError("An unexpected interval type (perhaps there's a bug or pending work here) tried "
                                 "to round a timestamp")



from ...pricing import Candle
from .side_pluckers import SidePlucker
from . import map


class CandlePlucker:
    """
    A CandlePlucker expects a source having dtype == Candle,
      and its implementation of __getitem__ wraps the call
      to the wrapped source, and plucks the specified field
      from each candle. The source must actually be a
      SidePlucker instance.
    """

    def __init__(self, source, component='end'):
        if not isinstance(source, SidePlucker) or source.dtype != Candle:
            raise TypeError("The source must be a SidePlucker, and Candle-based")
        if component not in Candle.__slots__:
            raise ValueError("For a candle-typed source frame, the component argument must be among (start, "
                             "end, min, max). By default, it will be 'end' (standing for the end price of the "
                             "candle)")
        self._component = component
        self._source = source

    @property
    def component(self):
        return self._component

    def _pluck(self, element):
        return [getattr(element[0], self._component)]

    def __getitem__(self, item):
        return map(self._source, item, self._pluck, int)

    def __len__(self):
        return len(self._source)

    @property
    def initial(self):
        """
        Returns a pluck on the initial value of the underlying source.
        """

        if self._source.initial is None:
            return None
        else:
            return getattr(self._source.initial, self._component)



class IdentityMapper:
    """
    Identity mappers return the values as unchanged.

    Their only purpose is either debugging or as a readonly proxy.
    """

    def __init__(self, parent):
        self._parent = parent

    def __getitem__(self, item):
        return self._parent[item]

    def __len__(self):
        return len(self._parent)



from numpy import float64, nan, isnan, vstack, ones
from ...utils.mappers.identity_mappers import IdentityMapper
from ...growing_arrays import GrowingArray


class ScoredMixin:
    """
    Adds capability to track its own score (a measure of performance)
      while running its lifecycle. In the end, the scoring is a float
      value in the range starting at 0 (shit) and ending at 1 (perfect),
      or NaN if a score is not available.

    Indicators implementing this trait/mixin will have its own growing
      array holding the scores for each time. Ideally, scoring involves
      several factors that evolve in time, but ultimately each score
      value does not depend on former values (unless a metric implies
      such behaviour).

    Implementors will have access to the `_scores` protected member to
      put a score for a given time, and a public `get_score` member to
      get the score for certain time. When putting a score, all the
      times before the score time will have a NaN value.
    """

    def __init__(self):
        self._scores = GrowingArray(float64, nan, 3600, 1)

    def get_score(self, time):
        """
        Gets the score(s) from a specific time instant or slice.
        Negative indices are not supported and are silently clamped.
        Specifying a step != 1 is not supported and will raise an
          exception.

        Retrieving a single entry before the reference origin (0),
          will return NaN. For slices, no values will exist there,
          and nothing will be included from the past, as NaNs like
          it occurs in the single case.

        Retrieving a slice in the future will consist of the last
          score value available... repeated for the whole size of
          the slice.

        Retrieving a slice in the present will be no different.

        Retrieving a slice that is partially in the present and
          partially in the future will imply getting actual values
          first, and getting repeated values last.
        :param time: The time instant or slice to retrieve.
        :return:
        """
        size = len(self._scores)
        if isinstance(time, int):
            # Two distinct cases live here:
            # - NaN will be returned if there are no items
            #   or the requested time index is negative.
            # - A defined value will be returned for positive
            #   times, returning the last available scoring
            #   if the index is out of bounds.
            if size == 0 or time < 0:
                return nan
            else:
                time = min(time, size - 1)
                return self._scores[time][:, 0]
        elif isinstance(time, slice):
            # First, slices will be preprocessed to only
            #   allow step=1 (or None), start < stop and
            #   both being integers (by default, start
            #   is 0 and stop is the array size). Both
            #   start and stop are clamped to disallow
            #   them to be negative values.
            start, stop, step = time.start, time.stop, time.step
            if not (step is None or step == 1):
                raise ValueError("Cannot slice the scores with a non-1 step")
            if start is None:
                start = 0
            if stop is None:
                stop = size
            start = max(0, start)
            stop = max(0, stop)
            if not (isinstance(start, int) and isinstance(stop, int)) or stop < start:
                raise ValueError("Slice's start and stop values must both be integer "
                                 "values and the start index must be <= the stop index")
            # The different scenarios will occur now:
            # - On size == 0, there is no available score.
            #   In this case an array of NaN values must
            #   be returned, with size == stop - start.
            # - On size <= start <= stop, an array will be
            #   returned but in this time it is filled with
            #   the last available score (the one at size
            #   - 1).
            # - On start <= stop <= size, an array of actual
            #   scores will be returned.
            # - The last case is start < size < stop. This
            #   case has both actual values and the filling
            #   with the last value, in two chunks:
            #   - [start:size]: actual values.
            #   - [size:stop]: fill values.
            if size == 0:
                return ones((stop - start,)) * nan
            elif start >= size:
                return ones((stop - start,)) * self._scores[size - 1][:, 0]
            elif stop <= size:
                return self._scores[start:stop][:, 0]
            else:
                actual = self._scores[start:size][:, 0]
                repeated = ones((stop - size,)) * self._scores[size - 1][:, 0]
                return vstack((actual, repeated))
        else:
            raise TypeError("Invalid item type while getting the scores from a scored "
                            "object: only int or slices are supported")


class EvolvingMetricScoredMixin(ScoredMixin):
    """
    This particular scored mixin provides a new protected member that
      will allow us to report values and compute scores.

    This works by also having a chain of metrics that must be setup in
      the _setup method, including a computation of the "final" metric
      (which will be stored in the _scores array). Details are as follows:

      1. A "zero" metric will be registered, and will be the "performance".
         Values to that metric are added on each call to this method:

         self._report_performance(time, value).

      2. Chained metrics can be added on setup by invoking this method:

         def _setup(self):
             i1 = self.metric(0, lambda time, p0: ...)
             # i1 will be == 1
             i2 = self.metric(0, i1, lambda time, p0, p1: ...)
             # i2 will be == 2

         The signature of each function must include an amount of parameters
         matching the count of dependent metrics indices, and those must not
         refer any yet-non-existing metric. The time may either be a scalar
         or a vector, and numpy can help us in both cases. The result must
         have the appropriate type and shape, depending on what the time
         argument actually is.

      3. In the same setup, the final call (this means: no more calls can be
         done to the setup object) looks like this:

             self.result_metric(0, 1, 3, lambda time, p0, p1, p3: ...)

         Which means: to compute a result for the score for the given time(s),
         just involve the metrics 0, 1, and 3 and compute something from them.
    """

    def __init__(self):
        """
        Initializes this indicator by running a setup, validating it,
          and preparing to run according to the required metrics and
          scoring mechanism.
        """

        self._can_setup = False
        super().__init__()
        self._metrics = []
        self._next_metric_index = 1
        self._time_array = GrowingArray(int, nan)
        self._metric_arrays = [
            GrowingArray(float, nan)
        ]
        self._can_setup = True
        self._result_metric = self._make_metric(*self._setup())
        self._can_setup = False

    def _setup(self):
        """
        This method MUST be implemented by subclasses.

        It must consist of several calls to _add_metric, and
          return a pair of (dependencies, callback) with the
          same signature as in the _add_metric arguments.
          Such function will determine the "result metric",
          which ideally should return a value between 0 and
          1 for each execution.
        """

        raise NotImplemented

    def _make_metric(self, dependencies, implementation):
        """
        Given a pair of dependencies and implementation, it creates a
          function taking just an index, and returning the value of a
          single metric. The function passed as implementation must
          take arguments: one for the integer (time) index, one for
          the internal time array (the instant when the scoring is
          occurring), and an arbitrary list of arguments being the
          dependencies. Such... list... must be of the same length
          of the deps array.
        :param dependencies: Dependencies indices.
        :param implementation: Metric implementation.
        :return: A single-parameter function that performs the
          logic by taking an index and the required arrays from
          the metric arrays.
        """

        arrays = [IdentityMapper(self._metric_arrays[dep]) for dep in dependencies]
        return lambda index: implementation(index, self._time_array, *arrays)

    def _add_metric(self, dependencies, implementation):
        """
        Creates and adds an internal metric. The implementation function
          must expect two arguments (time index, time array) and an arbitrary
          list of parameters being the dependencies: its size must be equal
          to the size of the dependencies iterable.
        :param dependencies: Dependencies indices.
        :param implementation: Metric implementation.
        :return: The index of the just-created metric.
        """
        if not self._can_setup:
            raise RuntimeError("Metrics can only be setup inside a _setup() call")

        idx = self._next_metric_index
        if any(dep for dep in dependencies if dep < 0 or dep >= idx):
            raise IndexError("The metric being added references dependencies: {0}, "
                             "but this indicator only has dependencies from 0 to "
                             "{1}".format(dependencies, idx - 1))

        self._metric_arrays.append(GrowingArray(float64, nan))
        self._metrics.append(self._make_metric(dependencies, implementation))
        self._next_metric_index += 1
        return idx

    def _performance_report(self, time, value):
        """
        Reports the performance at certain time. This function must be
          invoked by children indicators that want to assess their own
          performance to others.
        :param time: The time to set a performance value for.
        :param value: The performance value to set.
        """

        # NaN values are not accepted in either time or value.
        if isnan(time) or isnan(value):
            raise ValueError("Time and value must be actual numbers - not NaN")

        # The size of the time array serves as the next index to insert
        #   in the growing arrays.
        size = len(self._time_array)

        # Prevents that performance data is added for time instants that
        #   are equal or lower to the last time already added.
        if size and self._time_array[size - 1][0] >= time:
            raise ValueError("Cannot tell the performance for a time index earlier "
                             "or equal to the already reported ones")

        # The first step is to fill the time value, and the performance
        #   value, in the respective arrays: the time array and the first
        #   metric array, in both cases at the position given at the size
        #   variable.
        self._time_array[size] = time
        self._metric_arrays[0][size] = value

        # Then, the metric implementations and the metric arrays are both
        #   iterated as a pair (except for metric array 0), and setting
        #   the current index (the same size variable) to the value returned
        #   by the respective metric(s).
        for impl, array in zip(self._metrics, self._metric_arrays[1:]):
            array[size] = impl(size)

        # Setting the result involves running the result metric... but in
        #   this case, not in a metric array but in the score array, using
        #   as index the time value instead of the size of the former arrays.
        # Middle -unfilled- values will be filled to the -formerly- last score,
        #   reflecting that "nothing changed in the meantime".
        result = self._result_metric(size)
        self._scores[time - 1] = self._scores[len(self._scores) - 1]
        self._scores[time] = result



from datetime import timedelta


class DiscreteTimeDomain:
    """
    A discrete time domain is just a reference to a starting point,
    followed by a length and conditioned by a "bin size" (in terms
    of a time interval), and then having the "next" point. Both
    the start and next points are datetime objects, and the next
    point is the immediately next time point to arrive, typically.
    """

    def stamp_for(self, index):
        return self._get_timestamp() + timedelta(seconds=index * int(self._get_interval()))

    def index_for(self, stamp):
        return int((stamp - self._get_timestamp()).total_seconds()) // int(self._get_interval())

    @property
    def interval(self):
        """
        The interval size for this source. Digests must use BIGGER intervals in order to be able to
          connect to this source.
        """

        return self._get_interval()

    def _get_interval(self):
        """
        Abstract method that returns the interval to use.
        """

        raise NotImplemented

    @property
    def timestamp(self):
        """
        Stands for the initial timestamp (the one corresponding to the
        zero index) of this timelapse.
        :return: The initial timestamp.
        """

        return self._get_timestamp()

    def _get_timestamp(self):
        """
        Abstract method that returns the reference timestamp to use.
        """

        raise NotImplemented

    @property
    def next_timestamp(self):
        """
        Stands for the post-final timestamp / next timestamp (the one
        corresponding to the next stamp to use when adding new data).
        :return: The post-final / next timestamp.
        """

        return self.stamp_for(len(self))

    def __len__(self):
        """
        Computes the length of the current time space. This length
        can evolve: increment with new samples.
        """

        raise NotImplemented



import warnings
from datetime import date, datetime
from numpy import ndarray, uint64
from .timelapses import Timelapse
from .events import Event
from .pricing import StandardizedPrice, Candle


INTERPOLATION_WARNING_THRESHOLD = 30


class Source(Timelapse):
    """
    Source frames are the origin of the data. Internally, they are organized as a sequence of indexed prices
      or candles (depending on the required type: standardized price or candle).

    A type, an initial timestamp (e.g. day of activity) and an interval type is needed for the source frame
      know its context of work. After that, data may be added (either to the next available index or a given
      index that must be at least greater to the next one) and, if discontinuous, it will cause a kind of
      padding or interpolation in the frame data.

    When data is added, it will refresh two types of related (derived) frames:
      - Indicators of this frame.
      - Linked frames (to the current frame - they digest the data).
    """

    BID = 0  # Used to retrieve only the BID side of the source's data.
    ASK = 1  # Used to retrieve only the ASK side of the source's data.
    BOTH = 2  # Used to retrieve both sides. Each price will be retrieved as (bid, ask).

    class InterpolationWarning(Warning):
        pass

    def __init__(self, dtype, stamp, interval, initial_bid=None, initial_ask=None):
        """
        Creates a source frame with certain type, initial timestamp, an interval and an initial value.
        :param dtype: Either standardized price or candle.
        :param stamp: The initial time stamp. It will correspond to the sequence index 0.
        :param interval: The required interval.
        :param initial_bid: The initial bid for this frame. Usually being dragged from previous period.
          It must NOT be null if we don't plan to explicitly provide bid data for index 0.
        :param initial_ask: The initial ask for this frame. Usually being dragged from previous period.
          It must NOT be null if we don't plan to explicitly provide ask data for index 0.
        """

        if not interval.allowed_as_source():
            raise ValueError('The given interval is not allowed as source frame interval: %s' % interval)
        if dtype not in (StandardizedPrice, Candle):
            raise ValueError('The source frame type must be either pricing.Candle or pricing.StandardizedPrice')
        if initial_bid is not None:
            if dtype == StandardizedPrice and not isinstance(initial_bid, int):
                raise TypeError("For pricing.StandardizedPrice type, the initial bid value must be integer")
            elif dtype == Candle and not isinstance(initial_bid, Candle):
                raise TypeError("For pricing.Candle type, the initial bid value must be a candle instance")
        if initial_ask is not None:
            if dtype == StandardizedPrice and not isinstance(initial_ask, int):
                raise TypeError("For pricing.StandardizedPrice type, the initial ask value must be integer")
            elif dtype == Candle and not isinstance(initial_ask, Candle):
                raise TypeError("For pricing.Candle type, the initial ask value must be a candle instance")
        Timelapse.__init__(self, dtype, None if dtype == Candle else 0, 3600, 2)
        self._timestamp = stamp
        self._interval = interval
        self._initial = (initial_bid, initial_ask)
        self._on_refresh_linked_sources = Event()
        self._linked_to = None
        self._linked_last_read_ubound = 0
        self._linked_relative_bin_size = 0

    def _get_timestamp(self):
        """
        Implements the timestamp property by returning the owned timestamp.
        """

        return self._timestamp

    def _get_interval(self):
        """
        Implements the interval property by returning the owned interval.
        """

        return self._interval

    @property
    def on_refresh_indicators(self):
        """
        Indicators will connect to this event to refresh themselves when more data is added.
        """

        return self._on_refresh_indicators

    @property
    def initial(self):
        """
        The initial value of this source frame. When provided, this value is the last
          price (either StandardizedPrice or Candle) from the last period.
        """

        return self._initial

    def _interpolate(self, previous_values, start, end):
        """
        Causes an interpolation of data in certain index range, and considering
          boundary values.

        IT IS ACTUALLY A BAD PRACTICE TO HAVE TO INTERPOLATE DATA AND
        WILL LEAD TO MISLEADING RESULTS. Perhaps in a future I'm updating
        this part, but even there, it is a bad practice.

        :param previous_values: The left-side (bid, ask) value.
        :param start: The start index.
        :param end: The end index (not included).
        """

        # We'll state a new reference system where the previous value is at 0,
        #   and the next value is at distance, and we'll fill values from 1 to
        #   distance - 1.
        distance = end - start + 1
        if isinstance(previous_values, (uint64, int, StandardizedPrice)):
            previous_values = (previous_values,)
        # Iterating from index 1 to index {distance-1}, not including it.
        if isinstance(previous_values[0], (int, uint64)):
            for index in range(0, distance - 1):
                self._data[start + index] = previous_values
        elif isinstance(previous_values[0], Candle):
            previous_values = tuple(v.end for v in previous_values)
            for index in range(0, distance - 1):
                self._data[start + index] = tuple(Candle.constant(v) for v in previous_values)

    def _check_input_matching_types(self, data):
        """
        Checks whether the input matches the requirement.
        :param data: The data to check.
        """

        if isinstance(data, tuple) and len(data) == 2:
            for v in data:
                if isinstance(v, Candle):
                    if self._data.dtype != Candle:
                        raise TypeError("Scalar input data = (bid, ask) must match the required type")
                elif isinstance(v, (int, StandardizedPrice)):
                    if self._data.dtype != StandardizedPrice:
                        raise TypeError("Scalar input data = (bid, ask) must match the required type")
        elif not isinstance(data, ndarray) or len(data.shape) != 2 or data.shape[1] != 2:
            raise TypeError("Invalid input data type")

    def _put_and_interpolate(self, push_index, push_data):
        """
        First, it will put the data appropriately. Then it will try an interpolation -if needed- of the data,
          considering the last index, the initial value of this frame, and the first element of the pushed data.

        In the end, the new data (including the interpolated, if the case) will start at index
          self._last_index+1 and will end at index (push_index + {push_data.size or 1}), not
          including that index.
        :param push_index: The index the data is being pushed at.
        :param push_data: The data being pushed.
        """

        # Preliminary: Get the last index of data already put in this frame.
        length = len(self)

        # First, performs the insertion.
        is_ndarray = isinstance(push_data, ndarray)
        if is_ndarray:
            self._data[push_index:push_index+push_data.shape[0]] = push_data[:]
        else:
            self._data[push_index] = push_data

        # Check whether we need to interpolate, and do it.
        left_side = self._initial if length == 0 else self._data[length - 1]
        needs_interpolation = push_index > length
        if needs_interpolation:
            if push_index - 1 - length > INTERPOLATION_WARNING_THRESHOLD:
                warnings.warn(self.InterpolationWarning("Data is being added at sparse times! It may produce "
                                                        "misleading interpolated data."))
            if length == 0 and left_side is None:
                raise RuntimeError("Cannot add data: interpolation is needed for the required index "
                                   "to push the data into, but an initial value was never set for "
                                   "this frame")
            # Performs the interpolation.
            self._interpolate(left_side, length, push_index)

    def link(self, source):
        """
        Links this frame to another source, so data can be updated from it into this frame's data.
        It is an error to link to a source with a different interval size or a LOWER date: frames can
          only link to source with a greater date (and back-fill positions in previous time stamps).
        Linking to a source will automatically unlink from the previous source, if any.
        :param source: The source to link to.
        """

        self.unlink()
        if self._data.dtype != Candle:
            raise ValueError("This frame does not use Candle type, so it cannot connect to a source")
        if source.timestamp < self._timestamp:
            raise ValueError("The date of the source attempted to link to is lower than this frame's date")
        if int(source.interval) > int(self.interval):
            raise ValueError("The source to link to must have a lower interval than this frame")
        if int(self.interval) % int(source.interval) != 0:
            raise ValueError("The source's interval is not an exact divisor of this frame's interval")
        if self.interval.round(source.timestamp) != source.timestamp:
            raise ValueError("The specified source has not a date that is aligned to this frame's interval")
        self._linked_to = source._on_refresh_linked_sources
        self._linked_to.register(self._on_linked_refresh)
        self._linked_relative_bin_size = int(self.interval) // int(source.interval)
        # Force the first refresh.
        self._on_linked_refresh(source, 0, len(source))

    def unlink(self):
        """
        Unlinks this frame from its currently linked source.
        :return:
        """

        self._linked_last_read_ubound = 0
        self._linked_relative_bin_size = 0
        if self._linked_to:
            self._linked_to.unregister(self._on_linked_refresh)
            self._linked_to = None

    def _make_candles(self, source_elements):
        """
        Makes a candle out of the given source elements, either by summarizing integers, or candles.
          It will make two parallel candles: one for the bid prices, and one for the ask prices.
        :param source_elements: The elements to merge into one candle.
        :return: The two candles with the merged elements.
        """

        candle_bid, candle_ask = None, None
        for source_element in source_elements:
            def merge(candle, element):
                if candle is None:
                    if isinstance(element, uint64):
                        return Candle.constant(element)
                    elif isinstance(element, Candle):
                        return element
                else:
                    return candle.merge(element)
            candle_bid = merge(candle_bid, source_element[0])
            candle_ask = merge(candle_ask, source_element[1])
        return candle_bid, candle_ask

    def _on_linked_refresh(self, source, start, end):
        """
        Handles an update from the linked source considering start date and boundaries.
        It is guaranteed that boundaries will be in the same scale of this frame, but
          it must also be taken into account the start date to use as offset for the
          start and end indices.
        :param source: The linked source.
        :param start: The start index, in the source, of the updated data.
        :param end: The end index (not including), in the source, of the updated data.
        """

        base_index = self.index_for(source.timestamp)
        start = min(start, self._linked_last_read_ubound)
        min_index = start // self._linked_relative_bin_size
        max_index = (end + self._linked_relative_bin_size - 1) // self._linked_relative_bin_size
        for digest_index in range(min_index, max_index):
            source_index = digest_index * self._linked_relative_bin_size
            # We use indices 0 because we know the underlying array is of size 1.
            # From the source, we get a chunk of the relative size. Either a linear
            # array of integers, or a linear array of candles.
            # Now, make the candle out of the elements
            candles = self._make_candles(source._data[source_index:source_index+self._linked_relative_bin_size][:])
            self._data[digest_index + base_index] = candles
        self._linked_last_read_ubound = max(self._linked_last_read_ubound, end)
        # Indicators, and linked frames, of this source frame must also be notified, like in push.
        self._on_refresh_indicators.trigger(self, min_index + base_index, max_index + base_index)
        self._on_refresh_linked_sources.trigger(self, min_index + base_index, max_index + base_index)

    def __getitem__(self, item):
        """
        Allows retrieving certain item  (scalar or slice) from the underlying data.
          In particular, sources allow a second item to be specified, constrained to
          0 / Source.BID, 1 / Source.ASK, or 2 / Source.BOTH. By default, if no second
          item is specified, it is assumed to be == 2 / Source.BOTH.

        :param item: The item to fetch, in one of these formats: int, slice, (int, side), (slice, side).
          The side is a value in (0, 1, 2) as specified before.
        :return:
        """

        if isinstance(item, tuple):
            if len(item) != 2:
                raise IndexError(item)
            else:
                index, side = item
                if side == 2:
                    return super().__getitem__(index)
                elif side == 1 or side == 0:
                    return super().__getitem__((index, slice(side, side+1)))
                else:
                    raise IndexError(item)
        else:
            return super().__getitem__(item)

    def push(self, data, index=None):
        """
        Adds data, either scalar of the expected type, or a data chunk of the expected type.
        Always, at given index.

        :param data: The data to add. Scalar data is an iterable of length=2 containing two
          scalar values. Array data is an array of shape (whatever, 2).
        :param index: The index to add the data at. By default, the next index. It is allowed
          to update old data, but think it twice: it MAY trigger an update of old data, not account
          for actual data re-interpolation, and cascade the changes forward in unpredictable ways,
          which may involve SEVERAL QUITE COSTLY COMPUTATIONS!!.
        :return:
        """

        if index is None:
            index = len(self)
        elif isinstance(index, (date, datetime)):
            index = self.index_for(index)
        self._check_input_matching_types(data)
        self._put_and_interpolate(index, data)
        # Arrays have length in their shape, while other elements have size=1.
        end = index + (1 if not isinstance(data, ndarray) else data.shape[0])
        self._on_refresh_linked_sources.trigger(self, index, end)
        self._on_refresh_indicators.trigger(self, index, end)



from numpy import nan
from ...sources import Source
from ...utils.mappers.smart_pluckers import smart_plucker
from ...utils.tail_runners import TailRunner
from .. import Indicator


class MovingMean(Indicator):
    """
    Moving mean indicators are seldom used directly, but as dependencies for other indicators.
    They compute the moving mean of tail size = T, which is computed as the sample mean of the
      source elements in range [I-T+1:I+1].

    They can be used in the following frame types:
    - Integer source frames.
    - Float frames (other indicators) of width=1 (it is an error if they have different width).

    The tail size must be an integer greater than 1.

    Finally, it can be specified to tell this indicator to store NaN instead of a moving mean if
      the index is lower than (tail size - 1).

    Two additional arguments: component and row. They are required depending on the given parent.
    """

    def __init__(self, parent, tail_size, nan_on_short_tail=True, side=Source.ASK, component='end', row=0):
        self._parent = smart_plucker(parent, side, component, row)
        self._nan_on_short_tail = bool(nan_on_short_tail)
        self._tail_runner = TailRunner(tail_size)
        Indicator.__init__(self, parent)

    @property
    def tail_runner(self):
        """
        The internal tail runner.
        """

        return self._tail_runner

    @property
    def parent(self):
        """
        The direct parent of this moving mean.
        """

        return self._parent

    def _update(self, start, end):
        """
        Updates the indices with the moving/tailed mean for -respectively- each index.
        :param start: The start index to update.
        :param end: The end index to update.
        """

        for idx, chunk, incomplete in self._tail_runner.tail_iterate(start, end, self._parent):
            if incomplete and self._nan_on_short_tail:
                self._data[idx] = nan
            else:
                self._data[idx] = chunk.sum() / self._tail_runner.tail_size



"""
Indicators are a different monster. They handle a growing array of width N, where
  N is a different constant, depending on the indicator class. While row indices
  refer different instants, column indices refer different features the indicators
  may have. All the cells in this (dynamic)xN array will be of float type, and each
  indicator will have to deal with that to the appropriate extent.

Indicators will depend, alternatively, on (or: "be created with"):
  - A source frame (directly).
  - Another indicator(s) (directly) and the underlying frame (indirectly). All the
      indicators MUST have the same interval type (directly or not).

Indicators cannot unlink that dependency [although they should be able to halt (stop
  updating), resume, and dispose (halt forever, and destroy the underlying data in
  an unrecoverable way)] and pick another one, since it gives the appropriate way of
  refreshing the data on other -dependent- indicators. They are tied to the indicators
  or source they are created from, and those indicators or source will forward data
  updates to them.

When an indicator is disposed, all the dependent indicators will also dispose.

A quite complex network of indicators may be used. For example: One single moving mean
  of the last 20 time slices / candles / instants can feed a moving variance of the last
  20 time slices / candles / instants (which contains both the variance and the stderr),
  which in turn can feed 5 different Bollinger Bands indicators.
"""


from numpy import float64, nan
from ..timelapses import Timelapse


class Indicator(Timelapse):
    """
    Base class for indicators. Indicators depend ultimately on broadcasters, but they have
      to have the same source. Inheritors should have a way to distinguish each provided
      dependency.
    """

    def __init__(self, *broadcasters):
        broadcasters = set(broadcasters)
        intervals = set(broadcaster.interval for broadcaster in broadcasters)
        if len(intervals) != 1:
            raise ValueError("Indicators must receive at least a source and/or several other indicators, "
                             "and they must have the same interval")
        Timelapse.__init__(self, float64, nan, 3600, self.width())
        self._interval = intervals.pop()
        self._timestamp = max(broadcaster.timestamp for broadcaster in broadcasters)
        self._max_requested_start = {broadcaster: 0 for broadcaster in broadcasters}
        self._max_requested_end = {broadcaster: 0 for broadcaster in broadcasters}
        self._disposed = False
        # Trigger first refresh
        for broadcaster in broadcasters:
            broadcaster.on_refresh_indicators.register(self._on_dependency_update)
        for broadcaster in broadcasters:
            self._on_dependency_update(broadcaster, 0, len(broadcaster))

    def _get_timestamp(self):
        """
        Implements the timestamp property by returning the owned timestamp.
        """

        return self._timestamp

    def _get_interval(self):
        """
        Implements the interval property by returning the owned interval.
        """

        return self._interval

    def width(self):
        """
        The width of this indicator's data.
        """

        return 1

    def disposed(self):
        """
        Tells whether the current indicator is disposed (i.e. it will not work anymore, and data cannot
          be retrieved from it).
        """

        return self._disposed

    def __getitem__(self, item):
        """
        Returns data from this indicator.
        :param item: The index or slice to retrieve data at.
        """

        if self._disposed:
            raise RuntimeError("Cannot retrieve indicator data because it is disposed")
        return super().__getitem__(item)

    def dispose(self):
        """
        Clears this indicator from its dependency and broadcasts this call
          towards dependent indicators.
        """

        if not self._disposed:
            self._disposed = True
            self._data = None
            self._broadcasters_read = None
            for broadcaster in self._broadcasters_read.keys():
                broadcaster.on_refresh_indicators.unregister(self._on_dependency_update)
            for callback, receiver in self._on_refresh_indicators:
                receiver.dispose()

    def _on_dependency_update(self, dependency, start, end):
        """
        Processes a data update event. Such event will first be triggered from the source
        :param dependency: The dependency being updated.
        :param start: The internal start index of the dependency being updated.
        :param end: The internal end index of the dependency being updated.
        """

        # The maximum requested read for a dependency is the topmost read index
        #   up to now. Among every maximum requested index, we will get the minimum.
        self._max_requested_end[dependency] = max(end, self._max_requested_end[dependency])
        minimum_requested_end = min(r for r in self._max_requested_end.values())
        # Now we must also consider the minimum of this last index, and the currently requested end.
        current_end = min(minimum_requested_end, end)

        # For the start index (which will be <= than the end index), we still
        #   collect the maximums between the last read and the current start for,
        #   and still get the minimum among them.
        self._max_requested_start[dependency] = max(start, self._max_requested_start[dependency])
        minimum_requested_start = min(r for r in self._max_requested_start.values())
        # Then we also compare, and get minimum of, the minimum requested start, and the current start.
        current_start = min(minimum_requested_start, start)

        # Once we have these indices, we can invoke to refresh the data.
        self._update(current_start, current_end)

        # Then trigger to refresh the indicators.
        self._on_refresh_indicators.trigger(self, current_start, current_end)

    def _update(self, start, end):
        """
        Performs the update. This method must be implemented, account for all the needed dependencies, and also
          accounting for the fact that data may be NaN!
        :param start: The start index being refreshed.
        :param end: The end index (not included) being refreshed.
        """

        raise NotImplemented



from numpy import zeros, full
from .support import chunked_slicing, fix_slicing, fix_input


class GrowingArray:
    """
    A growing array can grow, but never shrink. Every time new data needs to be added, the array will
      perhaps grow to cover the required indices. Several chunks (all the same size) will be created
      on demand if a large index or a large slice -in both cases: before the current end- is provided.

    When getting an item or slice, IndexError will be raised as usual if out of bounds. Aside from that,
      data will be gathered across several different chunks if needed.

    In any case, negative indices are NOT supported, and steps different to 1 in slices are neither.
    """

    def __init__(self, dtype, fill_value, chunk_size=3600, width=1):
        if not isinstance(chunk_size, int) or chunk_size < 4:
            raise ValueError("Chunk size cannot be lower than 60")
        if not isinstance(width, int) or width < 1:
            raise ValueError("Width cannot be lower than 1")
        self._chunks = []
        self._dtype = dtype
        self._chunk_size = chunk_size
        self._fill_value = fill_value
        self._width = width
        self._length = 0

    @property
    def dtype(self):
        return self._dtype

    @property
    def width(self):
        return self._width

    def __getitem__(self, item):
        """
        Gets an element, or a numpy array of elements, given the index or slice.
        :param item: The index or slice to get the value from.
        :return: A numpy array with the specified items, if slice, or a single element.
        """

        start, stop, column = fix_slicing(item, self._length)
        return self._gather(start, stop, column)

    def _gather(self, start, stop, column):
        """
        Gathers required data from chunk(s).
        :param start: The start index to start gathering from.
        :param stop: The stop index (not included) to stop gathering from. It will be None if only one
          single element (a number or a (1, width) vector) is being retrieved.
        :param column: The column index or slice to retrieve a subset of the columns instead.
          If None, all the columns will be retrieved.
        :return: The gathered data (a single element, or a numpy array).
        """

        if stop is None:
            chunk_index = start // self._chunk_size
            chunk_pos = start % self._chunk_size
            chunk = self._chunks[chunk_index][chunk_pos]
            return chunk[:] if column is None else chunk[column]
        else:
            if isinstance(column, int):
                column = slice(column, column + 1, 1)
            elif column is None:
                column = slice(0, self._width, 1)
            if isinstance(column, slice):
                width = len(range(*column.indices(self._width)))
            else:
                raise IndexError("The second index to retrieve, if specified, must be an integer or a slice")
            data = zeros((stop-start, width), dtype=self._dtype)
            chunkings = chunked_slicing(start, stop, self._chunk_size)
            for (data_start, data_stop), chunk, (chunk_start, chunk_stop) in chunkings:
                data[data_start:data_stop, :] = self._chunks[chunk][chunk_start:chunk_stop, column]
            return data

    def _allocate(self, stop):
        """
        Allocates new arrays as needed, when needed, if needed.
        :param stop: The requested stop index.
        """

        chunks_count = len(self._chunks)
        total_allocated = chunks_count * self._chunk_size
        if stop > total_allocated:
            new_bins = (stop + self._chunk_size - 1) // self._chunk_size - chunks_count
            for _ in range(0, new_bins):
                arr = full((self._chunk_size, self._width), dtype=self._dtype, fill_value=self._fill_value)
                self._chunks.append(arr)
        self._length = max(self._length, stop)

    def _fill(self, start, stop, data):
        """
        Fills chunk(s) contents with given data.
        :param start: The start index to start filling.
        :param stop: The stop index (not included) to stop filling. It will be None if only one
          single element is being set.
        :param data: The data to fill with.
        :return:
        """

        if stop is None:
            chunk_index = start // self._chunk_size
            chunk_pos = start % self._chunk_size
            self._chunks[chunk_index][chunk_pos] = data
        else:
            chunkings = chunked_slicing(start, stop, self._chunk_size)
            for (data_start, data_stop), chunk, (chunk_start, chunk_stop) in chunkings:
                self._chunks[chunk][chunk_start:chunk_stop] = data[data_start:data_stop]

    def __setitem__(self, key, value):
        """
        Sets an element, or a numpy array of elements, given the index or slice.
        Chunks may be created on demand, depending on the index or slice being set.
        :param key: The index or slice to set the value into.
        :param value: The value to set.
        """

        start, stop, _ = fix_slicing(key, None, False)
        if stop is not None:
            value = fix_input(key, self._width, stop - start, self._dtype, value)
            self._allocate(stop)
            if stop > start:
                self._fill(start, stop, value)
        else:
            value = fix_input(key, self._width, 1, self._dtype, value)
            self._allocate(start + 1)
            self._fill(start, stop, value)

    def __repr__(self):
        return "GrowingArray(%s)" % ', '.join(str(chunk) for chunk in self._chunks)

    def __str__(self):
        return repr(self)

    def __len__(self):
        return self._length



from numpy import array


def map(source, item, function, dtype):
    """
    Given a "source" (an arbitrary object supporting 1-dimensional
      slicing and returning an arbitrary numpy array), it gets an
      item (which may be using an individual index, or a slice) and
      applies a mapping function depending the case: item or slice,
      and returning depending on the case: individual mapped item,
      or a numpy array of mapped values.
    :param source: The source to get the item from.
    :param item: The item to get, which may be int or slice.
    :param function: The mapping function to apply.
    :param dtype: The dtype to give the resulting array.
    :return: The mapped value.
    """

    data = source[item]
    if isinstance(item, slice):
        mapped = list(function(data[idx]) for idx in range(data.shape[0]))
        return array(mapped, dtype=dtype)
    elif isinstance(item, int):
        return function(data)
    else:
        raise TypeError("Unsupported item index type. Expected: slice or int")



from setuptools import setup

setup(
    name='mistra',
    version='0.1.1',
    packages=['mistra.core',
              'mistra.core.growing_arrays',
              'mistra.core.indicators',
              'mistra.core.indicators.mixins',
              'mistra.core.indicators.stats',
              'mistra.core.utils',
              'mistra.core.utils.mappers'],
    url='',
    license='MIT',
    author='luismasuelli',
    author_email='luisfmasuelli@gmail.com',
    description='MISTRA (Market InSights / TRading Algorithms) provides core support to market '
                'timelapses and indicators management',
    python_requires='>=3.12',
    install_requires=['numpy>=2.1.3', 'requests>=2.22.0', 'scipy>=1.14.1']
)



from numpy import nan
from ...growing_arrays import GrowingArray


class PredictorMixin:
    """
    Predictor indicators are, in the end, one kind of "useful" indicators.
      These indicators aim to be precise regarding the information they
      provide: literally the future price at certain (future) instant.

    To do so, they contain their own growing array and a forecast size
      >= 1: A prediction made for time t will be stored at slot t+F,
      where F is the forecast size, also an integer value.

    To get one of those predicted values, just call get_predicted(time),
      which takes a future time or a future time slice (this means: t+F
      was already called for those instants). Specifying a time lower
      than F will return NaN, while a time greater than or equal to the
      length of the array will raise an IndexError, and the same will
      occur if the time is lower than 0. F is the forecast size.

    To get one of those predictions (the predicted values before they
      occur) just call get_predictions(time) instead, which is like the
      previous get_predicted functions but moving the indices by +F.

    To add a new prediction at index t, which will stored at index t+F,
      make the indicator call _predict(time, value).
    """

    def __init__(self, forecast_size=1):
        if not (isinstance(forecast_size, int) and forecast_size < 0):
            raise ValueError("The forecast size must be a non-negative integer")
        self._forecast_size = forecast_size
        self._predictions = GrowingArray(float, nan)

    def _predict(self, time, value):
        """
        Adds a new prediction in certain time. The time will be
          fixed to its forecast version by adding the forecast
          size. In that final index, the time will be stored.

        Existing times cannot be overridden: predictions must
          be made in strict futures.
        :param time: The time when the prediction is made.
        :param value: The predicted value.
        """

        final_index = time + self._forecast_size
        length = len(self._predictions)
        if final_index < length:
            raise IndexError("New predictions must be appended in new positions "
                             "- predictions cannot be overridden")
        elif final_index == length:
            self._predictions[final_index] = value
        else:
            self._predictions[final_index - 1] = nan
            self._predictions[final_index] = value

    def get_predicted(self, time):
        """
        Gets the predictions for a specific time, which
          may be an integer or a slice.
        :param time: The instant or slice to get the predictions
          at.
        :return: A scalar or array of predictions.
        """

        return self._predictions[time]

    def get_predictions(self, time):
        """
        Get the predictions made at certain time instant or slice.
        :param time: The instant or slice when the prediction(s)
          were made.
        :return: A scalar or array of predictions.
        """

        if isinstance(time, int):
            time += self._forecast_size
        elif isinstance(time, slice):
            start, stop = time.start, time.stop
            start = (start or 0) + self._forecast_size
            stop = (stop or len(self._predictions)) + self._forecast_size
            time = slice(start, stop, 1)
        return self._predictions[time]
